unit UMain;

{FLX Sudoku Solver

Flix Guillemot Juin 2009

www.flx.fr}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Spin, ComCtrls,  Menus;

type
  TSudokuCell = class(TObject)
  private
    FSelected: boolean;
    FCellValue: SmallInt;
    FCountHypo: SmallInt;
    procedure SetSelected(const Value: boolean);
    procedure SetCellValue(const Value: SmallInt);
    procedure DrawCellValue(AValue : SmallInt);
    procedure FlushHypo;
    procedure InitCell;
    procedure SetCountHypo(const Value: SmallInt);

  public

    HypoArray   : Array[1..9] of Boolean;
    BgDrawn     : boolean;
    CellImage   : TImage;
    SudokuOwner : TObject;
    X           : integer;
    Y           : integer;

    procedure   DrawCellHypothese;
    procedure   CellClickEvent(Sender : TObject);
    Constructor Create(ASudokuOwner : TObject ; Owner : TWinControl ; iCol , irow : integer);
    Destructor  Destroy; override;
    Function    GetCellRect(AOffSet : SmallInt) : TRect;

    property Selected  : boolean read FSelected write SetSelected;
    property CellValue : SmallInt read FCellValue write SetCellValue;
    property CountHypo : SmallInt read FCountHypo write SetCountHypo;

end;

type
  TCellDir = (dirLeft , dirRight , dirUp , dirDown);



Type
  TArray9      = Array[1..9] of integer;

Type
  TArray1SQUARE = Array[1..4] of integer;

Type
  TArraySquare = Array[1..9] of TArray1SQUARE;

type
  TSudokuArray = Array[1..9 , 1..9] of TSudokuCell;

type
  TSudokuSave = Array[1..9 , 1..9] of SmallInt;

type
  TSudokuGame = class(TObject)
  private
    FOptAutoSolve: boolean;
    FOptSearchN: boolean;
    FCountSolution: integer;
    FNBMaxHypo: integer;


    procedure SetOptSearchN(const Value: boolean);
    procedure SetOptAutoSolve(const Value: boolean);
    function NumberInLine(iCol , iRow , iNum : integer ; var X , Y : integer) : boolean;
    function NumberInCol(iCol , iRow ,  iNum : integer ; var X , Y : integer) : boolean;
    function NumberInSquare(iCol , iRow ,  iNum : integer ; var X , Y : integer) : boolean;
    function HypoAlone(iHypo , iCol , iRow  : integer ; Square : TArray1SQUARE) : boolean;
    procedure InitSudokuArray;
    function DropHypoLine(iHypo , iRow  : integer; Square : TArray1SQUARE) : boolean;
    function DropHypoCol(iHypo , iCol  : integer ;  Square : TArray1SQUARE) : boolean;
    function OnlyOneHypo(iCol , iRow : integer) : SmallInt;
    procedure SetCountSolution(const Value: integer);
    function GetSelectedCell : TSudokuCell;
    procedure SetNBMaxHypo(const Value: integer);
    procedure UpdateNBMaxHypo;

  public
    SudokuArray     : TSudokuArray;
    SudokuSave      : TSudokuSave;
    ArraySudokuSave : array of TSudokuSave;
    Owner           : TWinControl;
    DrawMode        : boolean;



    Constructor Create(AOwner : TWinControl);
    destructor  Destroy; override;

    procedure SetSudokuArray(ANum: SmallInt; x, y: integer);
    procedure FillHypotheses;
    procedure UpdateHypotheses(Auto : boolean);
    procedure FlushSudokuArray;
    
    function  GetCellAt(iCol , iRow : integer) : TSudokuCell;
    function  GetCellFormImage(AImage : TImage) : TSudokuCell;
    procedure UnSelectAll;
    procedure AffectValueCell(AValue : SmallInt);

    procedure  SolveHypotheses;
    function  FixNumbers : boolean;

    function IsInLine(AValue , iCol , iRow : SmallInt ; Square : TArray1SQUARE): boolean;
    function IsInCol(AValue , iCol , iRow : SmallInt ; Square : TArray1SQUARE) : boolean;
    function CheckAll(var X1 , Y1 , X2 , Y2 : integer) : boolean;
    function DeepSolve(StartCol , StartRow : integer ; bSearchN : boolean) : boolean;

    function TwoHeadsHypo(iCol , iRow : SmallInt ;  var iNum1 , iNum2 : integer) : boolean;
    function ThreeHeadsHypo(iCol, iRow: SmallInt ; var iNum1 , iNum2 , iNum3 : integer): boolean;
    procedure InitSudokuSave; overload;
    procedure InitSudokuSave(var ASudokuSave : TSudokuSave); overload;
    procedure Save; overload;
    procedure Save(var ASudokuSave : TSudokuSave); overload;
    procedure reload; overload;
    procedure reload(ASudokuSave : TSudokuSave); overload;
    procedure reload(AStringList : TStringList); overload;
    function Full : boolean;
    procedure Refresh;
    procedure AppendSudokuSave(ASudoku : TSudokuSave);
    procedure HideHypotheses;

    property  OptAutoSolve  : boolean read FOptAutoSolve write SetOptAutoSolve;
    property  OptSearchN    : boolean read FOptSearchN write SetOptSearchN;
    property  CountSolution : integer read FCountSolution write SetCountSolution;
    property  NBMaxHypo     : integer read FNBMaxHypo write SetNBMaxHypo;




  end;



type
  TFrmSudoku = class(TForm)
    MainMenu1: TMainMenu;
    Jeu1: TMenuItem;
    Initialiser1: TMenuItem;
    Enregistrer1: TMenuItem;
    Vrifier1: TMenuItem;
    rsoudre1: TMenuItem;
    Exemples1: TMenuItem;
    AIEscargot1: TMenuItem;
    Diabolique1: TMenuItem;
    MenuOptions: TMenuItem;
    Rsolutionautomatique1: TMenuItem;
    Recherchedensolutions1: TMenuItem;
    PnlSolutions: TPanel;
    GBSolutions: TGroupBox;
    CBSolutions: TComboBox;
    Quitter1: TMenuItem;
    PnlSudoku: TPanel;
    Enregistrer2: TMenuItem;
    Apropos1: TMenuItem;
    Apropos2: TMenuItem;
    StatusBar1: TStatusBar;
    Importerunegrille1: TMenuItem;
    OpenDialog1: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure Button3Click(Sender: TObject);

    
    procedure Button5Click(Sender: TObject);
    procedure Button6Click(Sender: TObject);
    procedure CbAutoClick(Sender: TObject);
    procedure BtnVerifClick(Sender: TObject);
    procedure CBSolutionsCloseUp(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Initialiser1Click(Sender: TObject);
    procedure Enregistrer1Click(Sender: TObject);
    procedure Vrifier1Click(Sender: TObject);
    procedure rsoudre1Click(Sender: TObject);
    procedure Diabolique1Click(Sender: TObject);
    procedure AIEscargot1Click(Sender: TObject);
    procedure Rsolutionautomatique1Click(Sender: TObject);
    procedure Recherchedensolutions1Click(Sender: TObject);
    procedure Quitter1Click(Sender: TObject);
    procedure Enregistrer2Click(Sender: TObject);
    procedure Apropos2Click(Sender: TObject);
    procedure Importerunegrille1Click(Sender: TObject);


  private

    { Dclarations prives }
    SudokuGame      : TSudokuGame;
    FAbort: boolean;


    procedure AffectValueCell(ANum : SmallInt);
    procedure InitSudokuGame;
    procedure SetAbort(const Value: boolean);
    procedure MainReload;
    procedure MainCheck;
    procedure MainSolve;
    procedure SetHeight(ASolutions : boolean);
    procedure LoadSudokuFile;
    procedure MoveSelectedCell(Direction : TCellDir);


  public
    { Dclarations publiques }
    property Abort : boolean read FAbort write SetAbort;

  end;




function GetCellColor(iCol , iRow : integer) : TColor;  

var
  FrmSudoku: TFrmSudoku;
  gABORT : boolean;

Const
  CstDefaultColWidth  = 64;
  CstDefaultRowHeight = 64;
  CstCellColor1       = $00FCCBAD;
  CstCellColor2       = $00FEF1E9;
  CstCellCelectColor  = ClLime;

  ArraySq1 : TArray1SQUARE = (1 , 3 , 1 , 3);
  ArraySq2 : TArray1SQUARE = (4 , 6 , 1 , 3);
  ArraySq3 : TArray1SQUARE = (7 , 9 , 1 , 3);
  ArraySq4 : TArray1SQUARE = (1 , 3 , 4 , 6);
  ArraySq5 : TArray1SQUARE = (4 , 6 , 4 , 6);
  ArraySq6 : TArray1SQUARE = (7 , 9 , 4 , 6);
  ArraySq7 : TArray1SQUARE = (1 , 3 , 7 , 9);
  ArraySq8 : TArray1SQUARE = (4 , 6 , 7 , 9);
  ArraySq9 : TArray1SQUARE = (7 , 9 , 7 , 9);

Const


  AIEscargot : TSudokuSave = ((1 , 0 , 0 , 0 , 0 , 6 , 3 , 0 , 0),
                              (0 , 3 , 0 , 0 , 1 , 0 , 0 , 4 , 0),
                              (0 , 0 , 9 , 5 , 0 , 0 , 0 , 0 , 7),
                              (0 , 0 , 6 , 3 , 0 , 0 , 0 , 0 , 0),
                              (0 , 2 , 0 , 0 , 8 , 0 , 0 , 0 , 0),
                              (7 , 0 , 0 , 0 , 0 , 4 , 0 , 0 , 0),
                              (0 , 0 , 5 , 9 , 0 , 0 , 0 , 0 , 3),
                              (9 , 0 , 0 , 0 , 0 , 0 , 1 , 0 , 0),
                              (0 , 8 , 0 , 0 , 2 , 0 , 0 , 7 , 0));



  Diabolique : TSudokuSave = ((5 , 0 , 0 , 0 , 0 , 0 , 8 , 6 , 0),
                              (3 , 8 , 0 , 0 , 0 , 0 , 0 , 0 , 0),
                              (0 , 0 , 0 , 0 , 9 , 0 , 0 , 0 , 4),
                              (0 , 0 , 0 , 9 , 0 , 0 , 7 , 1 , 0),
                              (0 , 2 , 0 , 0 , 0 , 0 , 0 , 0 , 8),
                              (8 , 5 , 0 , 0 , 0 , 7 , 3 , 0 , 0),
                              (0 , 6 , 0 , 0 , 8 , 0 , 0 , 0 , 0),
                              (0 , 0 , 0 , 0 , 4 , 1 , 0 , 2 , 7),
                              (0 , 0 , 0 , 5 , 0 , 0 , 9 , 0 , 0));

implementation

uses DateUtils, UAbout;

{$R *.dfm}

procedure TSudokuCell.CellClickEvent(Sender : TObject);
begin
  TSudokuGame(SudokuOwner).UnSelectAll;

  Selected := true;
end;

procedure TSudokuCell.SetSelected(const Value: boolean);
begin
  FSelected := Value;

  DrawCellValue(CellValue);

end;

procedure TFrmSudoku.AffectValueCell(ANum: SmallInt);
begin
  SudokuGame.AffectValueCell(ANum);


  if SudokuGame.OptAutoSolve then
  begin
    SudokuGame.UpdateHypotheses(true);
  end;
end;



procedure TSudokuGame.AppendSudokuSave(ASudoku : TSudokuSave);
var
  NewTabLength : integer;
begin
  NewTabLength := length(ArraySudokuSave) + 1;
  SetLength(ArraySudokuSave , NewTabLength);
  ArraySudokuSave[high(ArraySudokuSave)] := ASudoku;
end;

procedure TFrmSudoku.AIEscargot1Click(Sender: TObject);
begin
  SudokuGame.reload(AIEscargot);
end;

procedure TFrmSudoku.Apropos2Click(Sender: TObject);
begin
  With TFrmAbout.create(self) do
  try
    Showmodal;
  finally
    free;
  end;
end;

procedure TFrmSudoku.BtnVerifClick(Sender: TObject);
begin
  MainCheck;
end;

procedure TSudokuGame.AffectValueCell(AValue: SmallInt);
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Affectation de la valeur  la cellule
      if SudokuArray[iCol , iRow].Selected then
      begin
        SudokuArray[iCol , iRow].CellValue := AValue;
      end;
    end;
  end;
end;

function TSudokuGame.CheckAll(var X1 , Y1 , X2 , Y2 : integer) : boolean;
var
  iCol , iRow : integer;
  iNum        : integer;
  iX , iY     : integer;
begin
  result := false;

  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      iNum := SudokuArray[iCol , iRow].CellValue;
      //Si il y a un chiffre dans la cellule
      if iNum <> 0 then
      begin
        //Condition 1 : la colonne qui contient le chiffre ne contient pas dj ce chiffre
        result := not(NumberInCol(iCol , iRow , iNum, iX , iY));

        //Si tjs ok
        if result then
        begin
          //Condition 2 : la ligne qui contient le chiffre ne contient pas dj ce chiffre
          result := not(NumberInLine(iCol , iRow , iNum , iX , iY));
        end;

        //Si tjs ok
        if result then
        begin
          //Condition 3 : le carr qui contient le chiffre ne contient pas dj ce chiffre
          result := not(NumberInSquare(iCol , iRow , iNum , iX , iY));
        end;
      end
      else
      begin
        //La valeur de la cellule est 0 --> il doit donc y avoir des hypothses dedans
        //Sinon le sudoku est faux
        if SudokuArray[iCol , iRow].CountHypo = 0 then
        begin
          result := false;
        end;
      end;
      if not result then
      begin
        X1 := iCol;
        Y1 := iRow;
        X2 := iX;
        Y2 := iY;
        exit;
      end;

    end;
  end;
end;



constructor TSudokuGame.create(AOwner : TWinControl);
begin
  inherited create;

  Owner     := AOwner;

  DrawMode  := true;

  InitSudokuArray;

end;

destructor TSudokuGame.destroy;
begin
  FlushSudokuArray;
  inherited;
end;


//Doit tre appel  chaque fois que l'on pose un nouveau chiffre
//et que l'on veut faire les dductions qui y sont lies
//En mode Auto, la rsolution d'hypothses entraine la dduction de chiffres poss
procedure TSudokuGame.UpdateHypotheses(Auto : boolean);
var
  NewNumberFixed : boolean;
begin

  NewNumberFixed := true;
  //Mise  jour des hypothses tant que de nouveaux chiffres sont poss
  While NewNumberFixed do
  begin
    //1 Pose les hypothses en fonction des chiffres dj en place
    FillHypotheses;
    //2 Rduit le nombre d'hypothses par dduction
    SolveHypotheses;

    NewNumberFixed := false;

    if Auto then
    begin
      //Peut entrainer la dduction de nouveaux chiffres en mode auto
      NewNumberFixed := FixNumbers;
    end;
  end;

  UpdateNBMaxHypo;

end;


procedure TSudokuGame.UpdateNBMaxHypo;
var
  iCol , iRow , iHypo ,  NbHypo  : integer;

begin
  FNbMaxHypo := 0;
  NbHypo     := 0;

  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Slt s'il n'y a pas de chiffre
      if SudokuArray[iCol , iRow].CellValue = 0 then
      begin
        NbHypo := 0;

        for iHypo := 1 to 9 do
        begin
          if SudokuArray[iCol , iRow].HypoArray[iHypo] then
          begin
            inc(NbHypo);
          end;
        end;
      end;

      if NbHypo > NbMaxHypo then
      begin
        FNbMaxHypo := NbHypo;
      end;

    end;

  end;

end;

procedure TSudokuGame.HideHypotheses;
var
  iCol , iRow , iHypo  : integer;

begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Slt s'il n'y a pas de chiffre
      if SudokuArray[iCol , iRow].CellValue = 0 then
      begin
        SudokuArray[iCol , iRow].InitCell;
        for iHypo := 1 to 9 do
        begin
          SudokuArray[iCol , iRow].HypoArray[iHypo] := false;
        end;
      end;
    end;

  end;
end;


procedure TSudokuGame.FillHypotheses;
var
  iCol , iRow  : integer;
  iNum         : integer;
  okHypo       : boolean;
  iX , iY      : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Slt s'il n'y a pas de chiffre
      if SudokuArray[iCol , iRow].CellValue = 0 then
      //Parcours des chiffres possibles de 1  9
      begin
        SudokuArray[iCol , iRow].CountHypo := 0;

        for iNum := 1 to 9 do
        begin
          //est-il possible de poser ce chiffre dans cette cellule ?

          //Condition 1 : la colonne qui contient le chiffre ne contient pas dj ce chiffre
          okHypo := not(NumberInCol(iCol , iRow , iNum , IX , iY));

          //Si l'hypothse est encore possible
          if okHypo then
          begin
            //Condition 2 : la ligne qui contient le chiffre ne contient pas dj ce chiffre
            okHypo := not(NumberInLine(iCol , iRow , iNum , iX , iY));
          end;

          //Si l'hypothse est encore possible
          if okHypo then
          begin
            //Condition 2 : le carr qui contient le chiffre ne contient pas dj ce chiffre
            okHypo := not(NumberInSquare(iCol , iRow , iNum , iX , iY));
          end;

          //Remplissage du tableau d'hypothses de la cellule du Sudoku
          SudokuArray[iCol , iRow].HypoArray[iNum] := okHypo;

          if okHypo then
          begin
             SudokuArray[iCol , iRow].CountHypo := SudokuArray[iCol , iRow].CountHypo + 1;
          end;   
        end;

        SudokuArray[iCol , iRow].DrawCellHypothese;

      end;
    end;

  end;
end;



function TSudokuGame.NumberInCol(iCol , iRow ,  iNum : integer ; var X , Y : integer) : boolean;
var
  CptLine : integer;
begin
  result := false;
  //Parcours des lignes de la ime colonne
  for CptLine := 1 to 9 do
  begin
    //Si la ligne n'est pas celle du chiffre observ
    if CptLine <> iRow then
    begin
      result := (SudokuArray[iCol][CptLine].CellValue = iNum);
      if result then
      begin
        X := iCol;
        Y := CptLine;
        exit;
      end;
    end;
  end;
end;

function TSudokuGame.NumberInLine(iCol , iRow ,  iNum : integer ; var X , Y : integer) : boolean;
var
  CptCol : integer;
begin
  result := false;
  //Parcours des colonnes de la ime ligne
  for CptCol := 1 to 9 do
  begin
    //Si la colonne n'est pas celle du chiffre observ
    if CptCol <> iCol then
    begin
      result := (SudokuArray[CptCol][iRow].CellValue = iNum);
    end;
    if result then
    begin
      X := CptCol;
      Y := iRow;
      exit;
    end;
  end;
end;

function TSudokuGame.NumberInSquare(iCol , iRow ,  iNum : integer ; var X , Y : integer) : boolean;
var
  iColMin, iColMax , iRowMin , iRowMax : integer;
  CptCol , CptRow : integer;
begin
  result := false;

  iRowMin := 0;
  iRowMax := 0;
  iColMin := 0;
  iColMax := 0;

  //Identifier dans quel carr on est pour dterminer les 4 bornes
  //de la boucle de recherche

  //Colonne
  if (iCol >= 1) and (iCol <= 3) then
  begin
    iColMin := 1;
    iColMax := 3;
  end;
  if (iCol >= 4) and (iCol <= 6) then
  begin
    iColMin := 4;
    iColMax := 6;
  end;
  if (iCol >= 7) and (iCol <= 9) then
  begin
    iColMin := 7;
    iColMax := 9;
  end;

  //Ligne
  if (iRow >= 1) and (iRow <= 3) then
  begin
    iRowMin := 1;
    iRowMax := 3;
  end;
  if (iRow >= 4) and (iRow <= 6) then
  begin
    iRowMin := 4;
    iRowMax := 6;
  end;
  if (iRow >= 7) and (iRow <= 9) then
  begin
    iRowMin := 7;
    iRowMax := 9;
  end;

  //Parcours des colonnes
  for CptCol := iColMin to iColMax do
  begin
    //Parcours des lignes
    for CptRow := iRowMin to iRowMax do
    begin
      //Si la colonne n'est pas celle du chiffre observ
      if (CptCol <> iCol) and (CptRow <> iRow) then
      begin
        result := (SudokuArray[CptCol , CptRow].CellValue = iNum);
      end;
      if result then
      begin
        X := CptCol;
        Y := CptRow;
        exit;
      end;
    end;
  end;

end;



function TSudokuGame.OnlyOneHypo(iCol, iRow: integer): SmallInt;
var
  iHypo   : SmallInt;
  CptHypo : SmallInt;
begin
  result  := 0;
  CptHypo := 0;
  for iHypo := 1 to 9 do
  begin
    if SudokuArray[iCol , iRow].HypoArray[iHypo] then
    begin
      inc(CptHypo);
      result := iHypo;
    end;
    //S'il y en a plus d'une
    if CptHypo > 1 then
    begin
      result := 0;
      exit;
    end;
  end;
end;

procedure TSudokuGame.Refresh;
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Destruction du TSudokuCell de la cellule
      SudokuArray[iCol , iRow].DrawCellValue(SudokuArray[iCol , iRow].CellValue);
    end;
  end;
end;

procedure TSudokuGame.reload(AStringList: TStringList);
var
  iCol , iRow : integer;
begin
  try

    OptAutoSolve := false;

    //parcours des colonnes
    for iCol := 1 to 9 do
    begin
      //parcours des lignes
      for iRow := 1 to 9 do
      begin
        SudokuArray[iCol , iRow].CellValue := 0;
        SudokuArray[iCol , iRow].CellValue := StrToInt(AStringList[iRow - 1][iCol]);
      end;
    end;

  except
    else
    begin
      MessageDlg('Format de fichier non valide !', mtError , [mbOK] , 0 );
    end;
  end;
end;

procedure TSudokuGame.FlushSudokuArray;
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Destruction du TSudokuCell de la cellule
      SudokuArray[iCol , iRow].free;
    end;
  end;
end;

function TSudokuGame.Full: boolean;
var
  iCol , iRow : integer;
begin
  result := true;
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Destruction du TSudokuCell de la cellule
      if SudokuArray[iCol , iRow].CellValue = 0 then
      begin
        result := false;
        exit;
      end;
    end;
  end;
end;

function TSudokuGame.GetCellAt(iCol, iRow : integer): TSudokuCell;
begin
  result := SudokuArray[iCol , iRow];
end;

function TSudokuGame.GetCellFormImage(AImage: TImage): TSudokuCell;
var
  iCol , iRow : integer;
begin
  result := nil;

  //Parcours des lignes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des colonnes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      if AImage = SudokuArray[iCol , iRow].CellImage then
      begin
        result := SudokuArray[iCol , iRow];
        exit;
      end;
    end;
  end;

end;

function TSudokuGame.GetSelectedCell: TSudokuCell;
var
  iCol , iRow : integer;
begin
  result := nil;
  //renvoie la cellule slectionne si il y en a une

  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Destruction du TSudokuCell de la cellule
      if SudokuArray[iCol , iRow].Selected then
      begin
        result := TSudokuCell(SudokuArray[iCol , iRow]);
        exit;
      end;
    end;
  end;
end;




procedure TFrmSudoku.Button3Click(Sender: TObject);
begin
  InitSudokuGame;
end;

procedure TFrmSudoku.Importerunegrille1Click(Sender: TObject);
begin
  LoadSudokuFile;
end;

procedure TFrmSudoku.Initialiser1Click(Sender: TObject);
begin
  InitSudokuGame;
end;

procedure TFrmSudoku.InitSudokuGame;
begin
  SudoKuGame.free;
  SudokuGame              := TSudokuGame.create(PnlSudoku);
  SudokuGame.OptAutoSolve := Rsolutionautomatique1.checked;
  SudokuGame.OptSearchN   := Recherchedensolutions1.checked;
  PnlSolutions.Visible    := false;
  SetHeight(false);
end;



procedure TFrmSudoku.LoadSudokuFile;
var
  StrList  : TStringList;
begin
  StrList := TStringList.create;

  //Charge la grille dans une TStringList
  if OpenDialog1.execute then
  try
    StrList.LoadFromFile(OpenDialog1.FileName);
    //Chargement de la grille  partir de la TStringList
    SudokuGame.reload(StrList);
  finally
   StrList.free;
  end;
end;

procedure TFrmSudoku.MainCheck;
var
  X1 , Y1 , X2 , Y2 : integer;
begin
  if not(SudokuGame.CheckAll(X1 , Y1 , X2 , Y2)) then
  begin
    Showmessage('pas OK : '+ '(' + inttostr(X1) + ' , ' + inttostr(Y1) + ')' + ' et ' + '(' + inttostr(X2) + ' , ' + inttostr(Y2) + ')' );
    SudokuGame.UnSelectAll;
    if (X1 in [1..9]) and (X2 in [1..9]) and (Y1 in [1..9]) and (Y2 in [1..9]) then
    begin
      SudokuGame.SudokuArray[X1 , Y1].Selected := true;
      SudokuGame.SudokuArray[X2 , Y2].Selected := true;
    end;
  end
  else
  begin
    Showmessage('OK');
  end;
end;

procedure TFrmSudoku.MainReload;
begin
  CBSolutions.visible := false;

  SudokuGame.Reload;

  if SudokuGame.OptAutoSolve then
  begin
    SudokuGame.UpdateHypotheses(true);
  end;
end;

procedure TFrmSudoku.MainSolve;
var
  i               : integer;
  DebutResolution : Int64;
  FinResolution   : Int64;
  Frequence       : Int64;
begin
  SudokuGame.CountSolution := 0;

  Screen.cursor := crHourGlass;

  //Pour ne pas dessiner pendant la recherche
  SudokuGame.DrawMode := false;
  SudokuGame.Save;
  gABORT := false;
  GBSolutions.visible := false;

  PnlSolutions.Visible := false;

  //initialisation du tableau de solutions
  SetLength(SudokuGame.ArraySudokuSave , 0);

  QueryPerformanceCounter(DebutResolution);


  //Mise  jour des hypothses de base
  SudokuGame.UpdateHypotheses(true);


  try
    //Si la rsolution vidente ne suffit pas : DeepSolve
    if not(SudokuGame.Full) then
    begin
      try
        SudokuGame.DeepSolve(1 , 1 , SudokuGame.OptSearchN);
      finally
        SudokuGame.DrawMode := true;
      end;

      QueryPerformanceCounter(FinResolution);
      QueryPerformanceFrequency(Frequence);
    end
    else
    begin
      QueryPerformanceCounter(FinResolution);
      QueryPerformanceFrequency(Frequence);

      //Si on n'a pas forc la rsolution avec solutions
      if not (SudokuGame.FOptSearchN) then
      begin
        SudokuGame.DrawMode := true;
        SudokuGame.UnSelectAll;
        exit;
      end;
    end;

    //gestion de l'affichage de ces solutions
    SudokuGame.DrawMode := true;

      
    if length(SudokuGame.ArraySudokuSave) > 1 then
    begin
      MessageDlg('Il y a plusieurs solutions.' , MtInformation , [MbOk] , 0);
      GBSolutions.visible := true;
      CBSolutions.visible := true;
      CBSolutions.clear;
      //Parcours des solutions
      for i := low(SudokuGame.ArraySudokuSave) to high(SudokuGame.ArraySudokuSave) do
      begin
        CBSolutions.Items.add('Solution N' + inttostr(i + 1));
      end;
      //Affichage des n solutions
      SetHeight(true);
      PnlSolutions.Visible := true;


      CBSolutions.ItemIndex := 0;
      SudokuGame.reload(SudokuGame.ArraySudokuSave[0]);
    end
    else
    //1 solution trouve : pas de liste de solutions
    begin
      if length(SudokuGame.ArraySudokuSave) = 1 then
      begin
        SudokuGame.reload(SudokuGame.ArraySudokuSave[0]);
      end;
    end;
    //Aucune solution trouve
    if (length(SudokuGame.ArraySudokuSave) = 0) and not(SudokuGame.Full) then
    begin
      SudokuGame.InitSudokuArray;
      SudokuGame.reload;
    end;


  finally

    if not SudokuGame.OptSearchN then
    begin
      StatusBar1.Panels[0].Text := 'Rsolu en ' +  Format('%0.4f ms', [(FinResolution-DebutResolution)*1000 / Frequence]);
    end
    else
    begin
      StatusBar1.panels[0].text := inttostr(SudokuGame.CountSolution) + ' solution(s) trouves en ' + Format('%0.4f ms', [(FinResolution-DebutResolution)*1000 / Frequence]);
    end;

    SudokuGame.UnSelectAll;
    Screen.cursor := crDefault;

  end;
end;

procedure TFrmSudoku.MoveSelectedCell(Direction: TCellDir);
var
  CurrentCell : TSudokuCell;
  X , Y       : integer;
begin
  CurrentCell := SudokuGame.GetSelectedCell;

  if CurrentCell = nil then
  begin
    SudokuGame.SudokuArray[1 , 1].Selected := true;
    exit;
  end;

  X := SudokuGame.GetSelectedCell.X;
  Y := SudokuGame.GetSelectedCell.Y;

  case Direction of
    dirLeft  : begin
                 if X = 1 then
                 begin
                   exit;
                 end;

                 CurrentCell.Selected := false;

                 CurrentCell          := SudokuGame.SudokuArray[X - 1 , Y];
                 CurrentCell.Selected := true;

               end;
    dirRight : begin
                 if X = 9 then
                 begin
                   exit;
                 end;

                 CurrentCell.Selected := false;

                 CurrentCell          := SudokuGame.SudokuArray[X + 1 , Y];
                 CurrentCell.Selected := true;
               end;
    dirUp    : begin
                 if Y = 1 then
                 begin
                   exit;
                 end;

                 CurrentCell.Selected := false;

                 CurrentCell          := SudokuGame.SudokuArray[X  , Y - 1];
                 CurrentCell.Selected := true;
               end;
    dirDown  : begin
                 if Y = 9 then
                 begin
                   exit;
                 end;

                 CurrentCell.Selected := false;

                 CurrentCell          := SudokuGame.SudokuArray[X  , Y + 1];
                 CurrentCell.Selected := true;
               end;
  end;





end;

procedure TFrmSudoku.Quitter1Click(Sender: TObject);
begin
  Self.close;
end;

procedure TFrmSudoku.Recherchedensolutions1Click(Sender: TObject);
begin
  (Sender as TMenuItem).checked := not((Sender as TMenuItem).checked);
  SudokuGame.OptSearchN         := (Sender as TMenuItem).checked;
end;

procedure TFrmSudoku.Rsolutionautomatique1Click(Sender: TObject);
begin
  (Sender as TMenuItem).checked := not((Sender as TMenuItem).checked);
  SudokuGame.OptAutoSolve       := (Sender as TMenuItem).checked;
end;

procedure TFrmSudoku.SetAbort(const Value: boolean);
begin
  FAbort := Value;
  gAbort := Value;
end;

procedure TFrmSudoku.SetHeight(ASolutions: boolean);
begin
  Autosize := false;

  if not ASolutions then
  begin
    self.height := (CstDefaultRowHeight * 9) + 16 + 47 + StatusBar1.height;
  end
  else
  begin
    self.height := (CstDefaultRowHeight * 9) + 16 + 47 + 50 + StatusBar1.height;
  end;

  Autosize := true;
end;


procedure TFrmSudoku.Vrifier1Click(Sender: TObject);
begin
  MainCheck
end;

procedure TFrmSudoku.Button5Click(Sender: TObject);
begin
  SudokuGame.Save;
end;

procedure TFrmSudoku.Button6Click(Sender: TObject);
begin
  MainReload;
end;

procedure TFrmSudoku.CbAutoClick(Sender: TObject);
begin
  if SudokuGame.OptAutoSolve then
  begin
    SudokuGame.UpdateHypotheses(true);
  end;
end;

procedure TFrmSudoku.CBSolutionsCloseUp(Sender: TObject);
begin
  SudokuGame.reload(SudokuGame.ArraySudokuSave[CBSolutions.ItemIndex]);
end;

procedure TFrmSudoku.Diabolique1Click(Sender: TObject);
begin
  SudokuGame.reload(Diabolique);
end;

procedure TFrmSudoku.Enregistrer1Click(Sender: TObject);
begin
  MainReload;
end;

procedure TFrmSudoku.Enregistrer2Click(Sender: TObject);
begin
  SudokuGame.Save;
end;

procedure TSudokuCell.InitCell;
var
  r            : TRect;
  ACellImage   : TImage;
  Canvas       : TCanvas;
begin
  if not(TSudokuGame(SudokuOwner).DrawMode) then
  begin
    exit;
  end;


  ACellImage := CellImage;
  Canvas     := ACellImage.Canvas;

  r          := GetCellRect(0);


  //fond du rectangle
  if not(Selected) then
  begin
    Canvas.brush.color := GetCellColor(X , Y);
  end
  else
  begin
    Canvas.brush.color := CstCellCelectColor;
  end;

  Canvas.Rectangle(r);


end;


procedure TSudokuCell.DrawCellHypothese;
var
  xMin , yMin  : integer;
  Canvas       : TCanvas;
  i            : integer;
  r            : TRect;
begin
  if not(TSudokuGame(SudokuOwner).DrawMode) then
  begin
    exit;
  end;


  InitCell;
  Canvas     := CellImage.Canvas;
  r          := GetCellRect(0);

  //Hypothses
  Canvas.Font.Size := 7;
  Canvas.Pen.Color := ClNavy;

  for i := low(HypoArray) to High(HypoArray) do
  begin
    xMin       := r.Left;
    yMin       := r.Top;

    if HypoArray[i] then
    begin
      //Positionnement du chiffre
      case i of

      1 : begin
            xMin := xMin + CstDefaultColWidth div 6;
            yMin := yMin + CstDefaultRowHeight div 6;
          end;
      2 : begin
            xMin := xMin + CstDefaultColWidth div 2 - 2;
            yMin := yMin + CstDefaultRowHeight div 6;
          end;
      3 : begin
            xMin := CstDefaultColWidth -  CstDefaultColWidth div 6 - 3;
            yMin := yMin + CstDefaultRowHeight div 6;
          end;
      4 : begin
            xMin := xMin + CstDefaultColWidth div 6;
            yMin := yMin + CstDefaultRowHeight div 2 - 2;
          end;
      5 : begin
            xMin := xMin + CstDefaultColWidth div 2 - 2;
            yMin := yMin + CstDefaultRowHeight div 2 - 3;
          end;
      6 : begin
            xMin := CstDefaultColWidth -  CstDefaultColWidth div 6 - 3;
            yMin := yMin + CstDefaultRowHeight div 2 - 2;
          end;
      7 : begin
            xMin := xMin + CstDefaultColWidth div 6;
            yMin := CstDefaultRowHeight - CstDefaultRowHeight div 6 - 4;
          end;
      8 : begin
            xMin := xMin + CstDefaultColWidth div 2 - 2;
            yMin := CstDefaultRowHeight - CstDefaultRowHeight div 6 - 4;
          end;
      9 : begin
            xMin := CstDefaultColWidth -  CstDefaultColWidth div 6 - 3;
            yMin := CstDefaultRowHeight - CstDefaultRowHeight div 6 - 4;
          end;
      end;

      xMin := xMin - 2;
      yMin := yMin - 2;
      Canvas.TextOut(xMin , yMin , inttostr(i));

    end;


  end;



end;




procedure TFrmSudoku.FormCreate(Sender: TObject);
begin
  SudokuGame := TSudokuGame.create(PnlSudoku);

  self.width := (CstDefaultColWidth * 9) +  16;
  SetHeight(false);

  PnlSolutions.Visible := false;
  Autosize := true;
end;

procedure TFrmSudoku.FormDestroy(Sender: TObject);
begin
  SudokuGame.free;

end;


procedure TFrmSudoku.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_DELETE then
     AffectValueCell(0);

       
  if Key = VK_LEFT then
  begin
    MoveSelectedCell(dirLeft);
  end;

  if Key = VK_RIGHT then
  begin
    MoveSelectedCell(dirRight);
  end;

  if Key = VK_UP then
  begin
    MoveSelectedCell(dirUp);
  end;

  if Key = VK_DOWN then
  begin
    MoveSelectedCell(dirDown);
  end;

end;

procedure TFrmSudoku.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = '1' then
     AffectValueCell(1);
  if Key = '2' then
     AffectValueCell(2);
  if Key = '3' then
     AffectValueCell(3);
  if Key = '4' then
     AffectValueCell(4);
  if Key = '5' then
     AffectValueCell(5);
  if Key = '6' then
     AffectValueCell(6);
  if Key = '7' then
     AffectValueCell(7);
  if Key = '8' then
     AffectValueCell(8);
  if Key = '9' then
     AffectValueCell(9);

end;

procedure TSudokuGame.InitSudokuSave;
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuSave) to high(SudokuSave) do
  begin
    //parcours des lignes
    for iRow := low(SudokuSave[iCol]) to high(SudokuSave[iCol]) do
    begin
      //Cration d'un objet TSudokuCell pour la cellule
      SudokuSave[iCol , iRow] := 0;

    end;

  end;

end;


procedure TSudokuGame.InitSudokuSave(var ASudokuSave : TSudokuSave);
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(ASudokuSave) to high(ASudokuSave) do
  begin
    //parcours des lignes
    for iRow := low(ASudokuSave[iCol]) to high(ASudokuSave[iCol]) do
    begin
      //Initialisation de la cellule
      ASudokuSave[iCol , iRow] := 0;

    end;

  end;

end;




procedure TSudokuGame.reload;
var
  iCol , iRow : integer;
begin
  InitSudokuArray;

  for iCol := low(SudokuSave) to high(SudokuSave) do
  begin
    //parcours des lignes
    for iRow := low(SudokuSave[iCol]) to high(SudokuSave[iCol]) do
    begin
      if SudokuSave[iCol , iRow] <> 0 then
      begin
        SudokuArray[iCol , iRow].CellValue := SudokuSave[iCol , iRow];
      end;
    end;
  end;
end;


procedure TSudokuGame.reload(ASudokuSave : TSudokuSave);
var
  iCol , iRow : integer;
begin
  OptAutoSolve := false;

  for iCol := low(ASudokuSave) to high(ASudokuSave) do
  begin
    //parcours des lignes
    for iRow := low(ASudokuSave[iCol]) to high(ASudokuSave[iCol]) do
    begin
      SudokuArray[iCol , iRow].CellValue := 0;
      SudokuArray[iCol , iRow].CellValue := ASudokuSave[iCol , iRow];
    end;
  end;
end;

procedure TSudokuGame.Save;
var
  iCol , iRow : integer;
begin
  InitSudokuSave;

  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      if SudokuArray[iCol , iRow].CellValue <> 0 then
      begin
        SudokuSave[iCol , iRow] := SudokuArray[iCol , iRow].CellValue;
      end;

    end;
  end;
end;


procedure TSudokuGame.Save(var ASudokuSave : TSudokuSave);
var
  iCol , iRow : integer;
begin
  //InitSudokuSave(ASudokuSave);

  //Parcours des colonnes
  for iCol := low(ASudokuSave) to high(ASudokuSave) do
  begin
    //parcours des lignes
    for iRow := low(ASudokuSave[iCol]) to high(ASudokuSave[iCol]) do
    begin
      if SudokuArray[iCol , iRow].CellValue <> 0 then
      begin
        ASudokuSave[iCol , iRow] := SudokuArray[iCol , iRow].CellValue;
      end
      else
      begin
        ASudokuSave[iCol , iRow] := 0;
      end;

    end;
  end;
end;




procedure TSudokuGame.InitSudokuArray;
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      if  SudokuArray[iCol , iRow] <> nil then
      begin
        SudokuArray[iCol , iRow].free;
      end;

      //Cration d'un objet TSudokuCell pour la cellule
      SudokuArray[iCol , iRow] := TSudokuCell.Create(Self , Owner , iCol , iRow);

    end;

  end;

end;

//Transforme les hypothses seules dans des cases en chiffres fixes
function TSudokuGame.FixNumbers : boolean;
var
  iCol , iRow : integer;
  LastHypo    : SmallInt;
begin
  result := false;
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //S'il n'y a plus qu'une seule hypothse l'hypothse devient un chiffre
      LastHypo := OnlyOneHypo(iCol , iRow);
      if (LastHypo > 0) then
      begin
        //Affectation de la valeur  cette cellule
        SudokuArray[iCol , iRow].CellValue := LastHypo;
        //pour signifier qu'au moins un nouveau chiffre a t pos
        result := true;
      end;
    end;
  end;

end;



procedure TSudokuGame.SetCountSolution(const Value: integer);
begin
  FCountSolution                  := Value;
  
end;

procedure TSudokuGame.SetNBMaxHypo(const Value: integer);
begin
  FNBMaxHypo := Value;
end;

procedure TSudokuGame.SetOptAutoSolve(const Value: boolean);
begin
  FOptAutoSolve := Value;

 // TFrmSudoku(Owner.owner).MenuOptions.Items[0].checked := Value;

  //Mise  jour des hypothses de base
  if (Value) then
  begin
    UpdateHypotheses(true);
  end
  else
  begin
    HideHypotheses;
  end;
end;

procedure TSudokuGame.SetOptSearchN(const Value: boolean);
begin
  FOptSearchN := Value;

  if not(Value) then
  begin
    TFrmSudoku(owner.owner).PnlSolutions.Visible := Value;
    TFrmSudoku(owner.owner).SetHeight(Value);
  end;
  //TFrmSudoku(Owner.owner).MenuOptions.Items[1].checked := Value;
end;

procedure TSudokuGame.SetSudokuArray(ANum: SmallInt; x, y: integer);
begin
  SudokuArray[x , y].CellValue := ANum;
end;






//procdure pour rsoudre lorsqu'il y a plusieurs combinaisons possibles
//true si jeu rsolu
function TSudokuGame.DeepSolve(StartCol , StartRow : integer ; bSearchN : boolean) : boolean;
var
  iCol , iRow : integer;
  iX1 , iY1   : integer;
  iX2 , iY2   : integer;
  iHypo       : integer;
  ArraySave   : TSudokuSave;
  ArraySaveOK : TSudokuSave;
  CountHypo   : integer;
begin
  result := false;

 // Application.ProcessMessages;
  if gABORT then
  begin
    exit;
  end;

  Screen.Cursor := CrHourGlass;


  //En entrant dans cette procdure, le jeu doit tre photographi
  // pour retrouver son tat avant de lui affecter une valeur
  Save(ArraySave);


  try

    result := false;

    //Tant que le jeu n'est pas rsolu
    While not (result) and not (Full) do
    begin
      //Parcours des colonnes
      for iCol := StartCol to high(SudokuArray)  do
      begin
        //parcours des lignes
        for iRow := StartRow to high(SudokuArray[iCol])  do
        begin
          CountHypo := SudokuArray[iCol , iRow].CountHypo;

          //Hypothse  n ttes (fonction de la profondeur)
          if (CountHypo <= NBMaxHypo) and (CountHypo > 0) then

          begin
            //Explorer les possibilits qui correspondent aux hypothses

            for iHypo := 1 to 9 do
            begin
              if (SudokuArray[iCol , iRow].HypoArray[iHypo]) then
              begin
              //  Application.ProcessMessages;
                if gABORT then
                begin
                  exit;
                end;

                {Avant d'affecter une quelconque valeur, retrouver le jeu tel qu'il tait en
                entrant dans la procdure
                }
                Reload(ArraySave);

                //essai
                SudokuArray[iCol , iRow].CellValue := iHypo;


                //Rsoud avec la situation ainsi produite
                UpdateHypotheses(true);

                //refresh;



                //le Sudoku est-il juste ?
                //Si non, inutile d'aller plus loin dans la recherche
                if CheckAll(iX1 , iY1 ,  iX2 , iY2) then
                begin
                  //Le sudoku est-il rempli ?
                  if Full then
                  begin
                    //Le sudoku est rsolu

                    //Sauvegarde de la solution de toute faon
                    Save(ArraySaveOK);
                    AppendSudokuSave(ArraySaveOK);
                    //Si on ne cherceh pas n solution mais une seule
                    if not(bSearchN) then
                    begin
                      result         := true;
                      CountSolution  := 1;
                     // DrawMode := true;
                      exit;
                    end
                    else
                    begin

                      inc(FCountSolution);
                      if FCountSolution = 1000000 then
                         gABORT := true;
                      Application.ProcessMessages;
                      TfrmSudoku(Owner.owner).StatusBar1.panels[0].text := inttostr(FCountSolution) + ' solutions trouves.';

                    end;

                  end
                  else
                  begin
                    //S'il n'est pas rempli, on continue d'explorer les hypothses en repartant du nouveau sudoku
                    //Comme cette cellule hypothse est transforme en chiffre, on ne passera pas dessus
                    result := DeepSolve(1 , 1 , bSearchN);

                    if not(result) and not(gABORT) then
                    begin
                      Reload(ArraySave);
                      UpdateHypotheses(true);
                    end;
                  end;
                end
                else
                begin
                  {Si le rsultat de l'hypothse donne un sudoku faux, il faut recharger
                  tout de suite le sudoku avant la pose de cette hypoyhse
                  sans quoi, (SudokuArray[iCol , iRow].HypoArray[iHypo]) ne sera plus jamais vrai
                  dans cette situation
                  }
                    
                  Reload(ArraySave);
                  UpdateHypotheses(true);

                end;

              end;
            end;
            {on a fait le tour des hypothses de cette case, et cela ne donne aucun rsultat
            donc ce n'est pas la peine de continuer dans cette voie car quelque soit la valeur
            que prendrait cette case parmi les hypothses possibles dans cette configuration
            ne fonctionnera pas, donc on sort immdiatement
            }
            exit;

          end;


        end;

      end;
      //on a fait le tour du sudoku
     exit;

    end;

  finally

    Screen.Cursor := CrDefault;
  end;
end;









procedure TSudokuGame.SolveHypotheses;
var

  ArraySquare : TArraySquare;
  iSquare     : integer;
  iColMin, iColMax , iRowMin , iRowMax : integer;
  iCol , iRow : integer;
  iHypo       : integer;
  CptHypo     : integer;
begin



  //les 9 carrs et leurs bornes
  ArraySquare[1] := ArraySq1;
  ArraySquare[2] := ArraySq2;
  ArraySquare[3] := ArraySq3;
  ArraySquare[4] := ArraySq4;
  ArraySquare[5] := ArraySq5;
  ArraySquare[6] := ArraySq6;
  ArraySquare[7] := ArraySq7;
  ArraySquare[8] := ArraySq8;
  ArraySquare[9] := ArraySq9;

  //parcours de chaque carr

  for iSquare := 1 to 9 do
  begin
    iColMin := ArraySquare[iSquare][1];
    iColMax := ArraySquare[iSquare][2];
    iRowMin := ArraySquare[iSquare][3];
    iRowMax := ArraySquare[iSquare][4];

    //Parcours des chiffres du carr

    //Parcours des colonnes du carr
    for iCol := iColMin to iColMax do
    begin
      //parcours des lignes
      for iRow := iRowMin to iRowMax do
      begin
         //Que les cellules  Hypothses
         if SudokuArray[iCol , iRow].CellValue = 0 then
         begin
           //parcours des 9 hypothses de la cellule
           for iHypo := 1 to 9 do
           begin
             //Si l'hypothse iHypo existe
             if SudokuArray[iCol , iRow].HypoArray[iHypo] then
             begin

               //On est sur une hypothse : est-elle la seule dans le carr ?
               if HypoAlone(iHypo , iCol , iRow , ArraySquare[iSquare]) then
               //Si oui, on limine toutes les autres hypothses dans la cellules
               //et on passe  la cellule suivante...
               begin
                 for CptHypo := 1 to 9 do
                 begin
                   if CptHypo <> iHypo then
                   begin
                     SudokuArray[iCol , iRow].HypoArray[CptHypo] := false;
                   end;
                 end;
                 //Rafrachissement du dessin des hypothses
                 SudokuArray[iCol , iRow].DrawCellHypothese;
                 break; //sort de la boucle des hypothses
               end
               else   //FLX TEMPO
               begin

                 //Si le chiffre et ses semblables sont en lignes
                 if IsInLine(iHypo , iCol , iRow , ArraySquare[iSquare]) then
                 begin
                  // Showmessage(inttostr(iHypo) + ' in line sur ' + inttostr(icol) + ' ; ' + inttostr(iRow));
                   //Alors limination des ventuelles hypothses sur la mme ligne des autres carrs
                   DropHypoLine(iHypo , iRow , ArraySquare[iSquare]);

                 end
                 else  //FLX TEMPO
                 begin

                   //Si le chiffre et ses semblables sont en colonnes
                   if IsInCol(iHypo , iCol , iRow , ArraySquare[iSquare]) then
                   begin
                   //  Showmessage(inttostr(iHypo) + ' en col sur ' + inttostr(icol) + ' ; ' + inttostr(iRow));
                     //Alors limination des ventuelles hypothses sur la mme colonnes
                     DropHypoCol(iHypo , iCol , ArraySquare[iSquare]);

                   end;

                 end;

               end;

             end;
           end;

         end;

      end;

    end;


  end;

end;



function TSudokuGame.ThreeHeadsHypo(iCol, iRow: SmallInt ; var iNum1 , iNum2 , iNum3 : integer): boolean;
var
  iHypo   : SmallInt;
  CptHypo : SmallInt;
begin
  iNum1 := 0;
  iNum2 := 0;
  iNum3 := 0;


  CptHypo := 0;
  for iHypo := 1 to 9 do
  begin
    if SudokuArray[iCol , iRow].HypoArray[iHypo] then
    begin
      if iNum1 = 0 then
      begin
        iNum1 :=iHypo;
      end
      else
      if iNum2 = 0 then
      begin
        iNum2 :=iHypo;
      end
      else
      begin
        iNum3 := iHypo;
      end;
      inc(CptHypo);
      if CptHypo > 3 then
      begin
        result := false;
        exit;
      end;
    end;
  end;

  result := (CptHypo = 2);

end;

function TSudokuGame.TwoHeadsHypo(iCol, iRow: SmallInt ; var iNum1 , iNum2 : integer): boolean;
var
  iHypo   : SmallInt;
  CptHypo : SmallInt;
begin
  iNum1 := 0;
  iNum2 := 0;

  CptHypo := 0;
  for iHypo := 1 to 9 do
  begin
    if SudokuArray[iCol , iRow].HypoArray[iHypo] then
    begin
      if iNum1 = 0 then
      begin
        iNum1 :=iHypo;
      end
      else
      begin
        iNum2 :=iHypo;
      end;
      inc(CptHypo);
      if CptHypo > 2 then
      begin
        result := false;
        exit;
      end;
    end;
  end;

  result := (CptHypo = 2);

end;

function TSudokuGame.DropHypoLine(iHypo , iRow  : integer ;  Square : TArray1SQUARE) : boolean;
var
  CptCol     : SmallInt;
begin
  result := false;
  for CptCol := 1 to 9 do
  begin
    //Eviter le carr en cours : en dehors des bornes
    if (CptCol < Square[1]) or (CptCol > Square[2]) then
    begin
       if SudokuArray[CptCol , iRow].HypoArray[iHypo] then
       begin
         SudokuArray[CptCol , iRow].HypoArray[iHypo] := false;
         SudokuArray[CptCol , iRow].DrawCellHypothese;
         result := true;
       end;
    end;

  end;
end;


function TSudokuGame.DropHypoCol(iHypo , iCol  : integer ;  Square : TArray1SQUARE) : boolean;
var
  CptRow : integer;
begin
  result := false;
  for CptRow := 1 to 9 do
  begin
    //Eviter le carr en cours : en dehors des bornes
    if (CptRow < Square[3]) or (CptRow > Square[4]) then
    begin
       if SudokuArray[iCol , CptRow].HypoArray[iHypo] then
       begin
         SudokuArray[iCol , CptRow].HypoArray[iHypo] := false;
         SudokuArray[iCol , CptRow].DrawCellHypothese;
         result := true;
       end;
    end;

  end;
end;

function TSudokuGame.HypoAlone(iHypo , iCol , iRow  : integer ; Square : TArray1SQUARE) : boolean;
var
  CptRow , CptCol : integer;
begin
  result := true;
  //Parcours des colonnes du carr
  for CptCol := Square[1] to Square[2] do
  begin
    //Parcours des lignes du carr
    for CptRow := Square[3] to Square[4] do
    begin
      //On vite la cellule de l'hypothse observe
      if not((CptCol = iCol) and (CptRow = iRow)) then
      begin
        //Si l'hypothse est sur la cellule
        if SudokuArray[CptCol , CptRow].HypoArray[iHypo] then
        begin
          result := false;
          exit;
        end;
      end;
    end;
  end;
end;

//fonction qui rpond  la question : un chiffre est il en ligne avec ses semblables ?
function TSudokuGame.IsInLine(AValue , iCol , iRow : SmallInt ; Square : TArray1SQUARE) : boolean;
var
  CptCol , CptRow : integer;
  SameInLine      : boolean;
begin
  //Par dfaut, oui
  result := true;

  //On est sur l'abscisse max du carr
  if  iCol = Square[2] then
  begin
    result := false;
    exit;
  end;

  SameInLine := false;

  //Question 1 : y-a-t il un semblable sur la mme ligne ?
  for CptCol := iCol + 1 to Square[2] do
  begin
    //Si l'hypothse pour ce chiffe existe sur cette cellule
    if SudokuArray[CptCol , iRow].HypoArray[AValue]  then
    begin
      //Oui, il y a un semblable
      SameInLine := true;
      break;
    end;
  end;

  //SSi il y a un semblable
  if not SameInLine then
  begin
    result := false;
    exit;
  end;


  //Contrle qu'il n'y a pas de semblables sur les autres lignes
  for CptRow := Square[3] to Square[4] do
  begin
    //Evite la ligne de dpart
    if CptRow <> iRow then
    begin
      for CptCol := Square[1] to Square[2] do
      begin
        //Si l'hypothse pour ce chiffe existe sur cette cellule
        if SudokuArray[CptCol , CptRow].HypoArray[AValue] then
        begin
          //Oui, il y a un semblable
          result := false;
          break;
        end;
      end;
    end;
  end;


end;


//fonction qui rpond  la question : un chiffre est il en ligne avec ses semblables ?
function TSudokuGame.IsInCol(AValue , iCol , iRow : SmallInt ; Square : TArray1SQUARE) : boolean;
var
  CptCol , CptRow : integer;
  SameInCol       : boolean;
begin
  //Par dfaut, oui
  result := true;

  //On est sur l'ordonne max du carr
  if  iRow = Square[4] then
  begin
    result := false;
    exit;
  end;

  SameInCol := false;

  //Question 1 : y-a-t il un semblable sur la mme colonne ?
  for CptRow := iRow + 1 to Square[4] do
  begin
    //Si l'hypothse pour ce chiffe existe sur cette cellule
    if SudokuArray[iCol , CptRow].HypoArray[AValue]  then
    begin
      //Oui, il y a un semblable
      SameInCol := true;
      break;
    end;
  end;

  //SSi il y a un semblable
  if not SameInCol then
  begin
    result := false;
    exit;
  end;


  //Contrle qu'il n'y a pas de semblables sur les autres colonnes
  for CptCol := Square[1] to Square[2] do
  begin
    //Evite la ligne de dpart
    if CptCol <> iCol then
    begin
      for CptRow := Square[3] to Square[4] do
      begin
        //Si l'hypothse pour ce chiffe existe sur cette cellule
        if SudokuArray[CptCol , CptRow].HypoArray[AValue] then
        begin
          //Oui, il y a un semblable
          result := false;
          break;
        end;
      end;
    end;
  end;


end;



procedure TSudokuGame.UnSelectAll;
var
  iCol , iRow : integer;
begin
  //Parcours des colonnes
  for iCol := low(SudokuArray) to high(SudokuArray) do
  begin
    //parcours des lignes
    for iRow := low(SudokuArray[iCol]) to high(SudokuArray[iCol]) do
    begin
      //Dselction
      SudokuArray[iCol , iRow].Selected := false;
    end;
  end;
end;

function GetCellColor(iCol , iRow : integer) : TColor;
begin
  //Couleur par dfaut
  result := CstCellColor2;

  //Carrs de gauche
  if (((iCol >= 1) and (iCol <= 3))
     and ((iRow >= 1) and (iRow <= 3))) or
      (((iCol >= 1) and (iCol <= 3))
     and ((iRow >= 7) and (iRow <= 9))) then
  begin
    Result := CstCellColor1;
  end;

  //Carr du milieu
  if (((iCol >= 4) and (iCol <= 6))
     and ((iRow >= 4) and (iRow <= 6))) then
  begin
    Result := CstCellColor1;
  end;

  //Carrs de droite
  if (((iCol >= 7) and (iCol <= 9))
     and ((iRow >= 1) and (iRow <= 3))) or
      (((iCol >= 7) and (iCol <= 9))
     and ((iRow >= 7) and (iRow <= 9))) then
  begin
    Result := CstCellColor1;
  end;

end;

{ TSudokuCell }



constructor TSudokuCell.Create(ASudokuOwner : TObject ; Owner : TWinControl ; iCol , iRow : integer);
var
  i      : integer;
  Canvas : TCanvas;
  r      : TRect;
begin
  inherited create;

  SudokuOwner        := ASudokuOwner;
  X                  := iCol;
  Y                  := iRow;




  CellImage          := TImage.create(Owner);
  CellImage.Parent   := Owner as TWincontrol;

  //Coordonnes de l'image  sa cration
  CellImage.Left     := CstDefaultColWidth * (iCol - 1) + iCol - 1;
  CellImage.Top      := CstDefaultRowHeight * (iRow - 1) + iRow - 1;

  //Taille de l'image  sa cration
  CellImage.Width    := CstDefaultColWidth;
  CellImage.Height   := CstDefaultRowHeight;

  CellImage.OnClick := CellClickEvent;

  if (TSudokuGame(SudokuOwner)).DrawMode then
  begin
    //Couleur de l'image
    Canvas             := CellImage.Canvas;

    r.Left             := 0;
    r.Right            := CellImage.width;
    r.Top              := 0;
    r.Bottom           := CellImage.Height;


    Canvas.Brush.color := GetCellColor(iCol , iRow);

    Canvas.Rectangle(r);
  end;

  //Initialisation de la valeur globale de la cellule  0
  CellValue := 0;

  //Initialisation du tableau d'hypothses relatif  la cellule
  for i := low(HypoArray) to High(HypoArray) do
  begin
    HypoArray[i] := false;
  end;

end;

destructor TSudokuCell.Destroy;
begin
  CellImage.free;
  inherited;
end;

function TSudokuCell.GetCellRect(AOffSet : SmallInt) : TRect;
begin
  result.Left             := 0 + AOffSet;
  result.Right            := CellImage.width - AOffSet;
  result.Top              := 0 + AOffSet;
  result.Bottom           := CellImage.Height - AOffSet;
end;

procedure TSudokuCell.SetCellValue(const Value: SmallInt);
begin
  FCellValue := Value;

  //dessin du chiffre
  if TSudokuGame(SudokuOwner).DrawMode then
  begin
    DrawCellValue(Value);
  end;


  if Value <> 0 then
  begin
    FlushHypo;
  end;

  //Mise  jour du nombre d'hypothses
  FCountHypo := 0;


end;

procedure TSudokuCell.SetCountHypo(const Value: SmallInt);
begin
  FCountHypo := Value;
end;

procedure TSudokuCell.DrawCellValue(AValue : SmallInt);
var
  r      : TRect;
  CanVas : TCanvas;
  sNum   : string;
begin
  if not(TSudokuGame(SudokuOwner)).DrawMode then
  begin
    exit;
  end;

  sNum             := inttostr(AValue);
  Canvas           := CellImage.Canvas;
  if AValue = 0 then
  begin
    r                := GetCellRect(0);
  end
  else
  begin
    r                := GetCellRect(1);
  end;

  if FSelected then
  begin
    Canvas.Brush.Color := CstCellCelectColor;
  end
  else
  begin
    Canvas.Brush.Color := GetCellColor(X , Y);
  end;

  Canvas.Rectangle(r);

  if AValue <> 0 then
  begin
    Canvas.Font.Size := 20;
    Canvas.TextRect(r , r.Left + 22 , r.Top + 15 , sNum);
  
  end
  else
  begin
    DrawCellHypothese;
  end;
end;

procedure TSudokuCell.FlushHypo;
var
  i : integer;
begin
  for i := 1 to 9 do
  begin
    HypoArray[i] := false;
  end;
end;



procedure TFrmSudoku.rsoudre1Click(Sender: TObject);
begin
  MainSolve;
end;

end.
